package nl.tue.astar.impl.cached;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.set.TLongSet;
import gnu.trove.set.hash.TLongHashSet;

import java.io.IOException;
import java.util.List;

import nl.tue.astar.AStarException;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.Delegate;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.State;
import nl.tue.astar.impl.fast.FastAStarThread;
import nl.tue.storage.impl.SkippableOutputStream;

public class CachedFastAStarThread<H extends Head, T extends FastLowerBoundTail>
		extends FastAStarThread<H, T> {

	public CachedFastAStarThread(Delegate<H, T> delegate,
			TObjectIntMap<H> head2int, List<State<H, T>> stateList,
			H initialHead, Trace trace, int maxStates) throws AStarException {
		super(delegate, head2int, stateList, initialHead, trace, maxStates);
	}

	@Override
	protected Record poll() {
		Record node = queue.poll();
		return node;
	}
	
	@Override
	@SuppressWarnings("unchecked")
	protected T computeNewTail(T tail, H newHead, int modelMove,
			int movedEvent, int activity) {
		T newTail = (T) tail.getNextTail(delegate, newHead, modelMove,
				movedEvent, activity);

		if (newTail.isExactEstimateKnown()) {
			for (AStarObserver observer : observers) {
				observer.estimateComputed(newHead);
			}
			computedEstimateCount++;
		}

		return newTail;
	}


	@Override
	protected void processMovesForRecord(Record rec, H head, T tail,
			int stopAt, double timeLimit, long endTime) throws AStarException {
		int storedEstimate = rec.getEstimatedRemainingCost();

		if (storedEstimate < 0) {
			// must be the initial state.
			// rec.setEstimatedRemainingCost(tailEstimate);

			considered.add(rec.getState());
			for (AStarObserver observer : observers) {
				observer.nodeVisited(rec);
			}
			super.processMovesForRecord(rec, head, tail, stopAt, timeLimit,
					endTime);
		} else {
			if (!tail.isExactEstimateKnown()) {
				tail.computeEstimate(delegate, head, storedEstimate);
				for (AStarObserver observer : observers) {
					observer.estimateComputed(head);
				}
				computedEstimateCount++;
			}

			int tailEstimate = tail.getEstimatedCosts(delegate, head);

			if (storedEstimate < tailEstimate) {
				// the tail's true estimate is higher than the fast one stored
				// earlier. Re-queue the record
				rec.setEstimatedRemainingCost(tailEstimate);

				if (queue.add(rec)) {
					queuedStateCount++;
				}
				poll--;
				// edge from oldRec to newRec traversed.
				for (AStarObserver observer : observers) {
					observer.edgeTraversed(rec, rec);
				}
			} else {
				considered.add(rec.getState());
				for (AStarObserver observer : observers) {
					observer.nodeVisited(rec);
				}
				super.processMovesForRecord(rec, head, tail, stopAt, timeLimit,
						endTime);
			}

		}
	}

}
